#define TK_EOL                             1
#define TK_ASSIGN                          2
#define TK_FORCE_ASSIGN                    3
#define TK_APPEND                          4
#define TK_LKEY                            5
#define TK_PLUS                            6
#define TK_STRING                          7
#define TK_INTEGER                         8
#define TK_LPARAN                          9
#define TK_RPARAN                         10
#define TK_COMMA                          11
#define TK_ARRAY_ASSIGN                   12
#define TK_GLOBAL                         13
#define TK_LCURLY                         14
#define TK_RCURLY                         15
#define TK_ELSE                           16
#define TK_DOLLAR                         17
#define TK_SRVVARNAME                     18
#define TK_LBRACKET                       19
#define TK_RBRACKET                       20
#define TK_EQ                             21
#define TK_MATCH                          22
#define TK_NE                             23
#define TK_NOMATCH                        24
#define TK_INCLUDE                        25
#define TK_INCLUDE_SHELL                  26
